@@ -1,6 +1,9 @@
 #!/usr/bin/env python
 # -*- coding: UTF8 -*-
 
+#@todo Research Fn
+#@todo Research optimizations
+
 from __future__ import with_statement
 
 import sys
@@ -25,6 +28,10 @@ IS_MAEMO = True
 
 
 def split_number(number):
+	if number == 0.0:
+		# Optimize the startup case
+		return "0.", "0"
+
 	try:
 		fractional, integer = math.modf(number)
 	except TypeError:
@@ -36,15 +43,16 @@ def split_number(number):
 		if "e+" in integerDisplay:
 			integerDisplay = number
 			fractionalDisplay = ""
-		elif "e-" in fractionalDisplay and 0.0 < integer:
-			integerDisplay = number
-			fractionalDisplay = ""
 		elif "e-" in fractionalDisplay:
-			integerDisplay = ""
-			fractionalDisplay = number
+			if 0.0 < integer:
+				integerDisplay = number
+				fractionalDisplay = ""
+			else:
+				integerDisplay = ""
+				fractionalDisplay = number
 		else:
-			integerDisplay = integerDisplay.split(".", 1)[0] + "."
-			fractionalDisplay = fractionalDisplay.rsplit(".", 1)[-1]
+			integerDisplay = integerDisplay[0:-2] + "."
+			fractionalDisplay = fractionalDisplay[2:]
 
 	return integerDisplay, fractionalDisplay
 
@@ -614,11 +622,13 @@ class QuickConvert(object):
 		self._favoritesWindow = None
 
 		self._inputUnitValue = QtGui.QLineEdit()
-		self._inputUnitValue.setInputMethodHints(QtCore.Qt.ImhPreferNumbers)
+		maeqt.mark_numbers_preferred(self._inputUnitValue)
 		self._inputUnitValue.textEdited.connect(self._on_value_edited)
 		self._inputUnitSymbol = QtGui.QLabel()
 
-		self._outputUnitValue = QtGui.QLabel()
+		self._outputUnitValue = QtGui.QLineEdit()
+		maeqt.mark_numbers_preferred(self._outputUnitValue)
+		self._outputUnitValue.textEdited.connect(self._on_output_value_edited)
 		self._outputUnitSymbol = QtGui.QLabel()
 
 		self._conversionLayout = QtGui.QHBoxLayout()
@@ -756,6 +766,39 @@ class QuickConvert(object):
 			self._window.showNormal()
 
 	def select_category(self, categoryName):
+		self._select_category(categoryName)
+
+		i = unit_data.UNIT_CATEGORIES.index(categoryName)
+		rootIndex = self._categoryView.rootIndex()
+		currentIndex = self._categoryView.model().index(i, 0, rootIndex)
+		self._categoryView.scrollTo(currentIndex)
+		self._categoryView.setItemSelected(self._categoryView.topLevelItem(i), True)
+
+		return self
+
+	def select_unit(self, name):
+		self.select_input(name)
+		return self
+
+	def select_input(self, name):
+		self._select_input(name)
+
+		i = self._unitNames.index(name)
+		rootIndex = self._inputView.rootIndex()
+		currentIndex = self._inputView.model().index(i, 0, rootIndex)
+		self._inputView.scrollTo(currentIndex)
+		self._inputView.setItemSelected(self._inputView.topLevelItem(i), True)
+
+	def select_output(self, name):
+		self._select_output(name)
+
+		i = self._unitNames.index(name)
+		rootIndex = self._outputView.rootIndex()
+		currentIndex = self._outputView.model().index(i, 0, rootIndex)
+		self._outputView.scrollTo(currentIndex)
+		self._outputView.setItemSelected(self._outputView.topLevelItem(i), True)
+
+	def _select_category(self, categoryName):
 		self._inputUnitName = ""
 		self._outputUnitName = ""
 		self._inputUnitValue.setText("")
@@ -772,8 +815,7 @@ class QuickConvert(object):
 		self._unitNames.sort()
 		for key in self._unitNames:
 			conversion, unit, description = unitData[key]
-			if not unit:
-				unit = key
+			unit = key
 
 			twi = QtGui.QTreeWidgetItem(self._inputView)
 			twi.setText(0, unit)
@@ -783,11 +825,6 @@ class QuickConvert(object):
 			twi.setText(0, unit)
 			twi.setText(1, key)
 
-		i = unit_data.UNIT_CATEGORIES.index(categoryName)
-		rootIndex = self._categoryView.rootIndex()
-		currentIndex = self._categoryView.model().index(i, 0, rootIndex)
-		self._categoryView.scrollTo(currentIndex)
-
 		defaultInputUnitName = self._app.get_recent_unit(categoryName)
 		if defaultInputUnitName:
 			self.select_input(defaultInputUnitName)
@@ -795,12 +832,7 @@ class QuickConvert(object):
 			assert defaultOutputUnitName
 			self.select_output(defaultOutputUnitName)
 
-		return self
-
-	def select_unit(self, name):
-		self.select_input(name)
-
-	def select_input(self, name):
+	def _select_input(self, name):
 		self._app.add_recent(self._categoryName, name)
 		self._inputUnitName = name
 
@@ -809,15 +841,10 @@ class QuickConvert(object):
 
 		self._inputUnitSymbol.setText(unit if unit else name)
 
-		i = self._unitNames.index(name)
-		rootIndex = self._inputView.rootIndex()
-		currentIndex = self._inputView.model().index(i, 0, rootIndex)
-		self._inputView.scrollTo(currentIndex)
-
 		if "" not in [self._categoryName, self._inputUnitName, self._outputUnitName]:
-			self._update_conversion()
+			self._update_output()
 
-	def select_output(self, name):
+	def _select_output(self, name):
 		# Add the output to recent but don't make things weird by making it the most recent
 		self._app.add_recent(self._categoryName, name)
 		self._app.add_recent(self._categoryName, self._inputUnitName)
@@ -828,13 +855,8 @@ class QuickConvert(object):
 
 		self._outputUnitSymbol.setText(unit if unit else name)
 
-		i = self._unitNames.index(name)
-		rootIndex = self._outputView.rootIndex()
-		currentIndex = self._outputView.model().index(i, 0, rootIndex)
-		self._outputView.scrollTo(currentIndex)
-
 		if "" not in [self._categoryName, self._inputUnitName, self._outputUnitName]:
-			self._update_conversion()
+			self._update_output()
 
 	def _sanitize_value(self, userEntry):
 		if self._categoryName == "Computer Numbers":
@@ -849,7 +871,7 @@ class QuickConvert(object):
 				value = float(userEntry)
 		return value
 
-	def _update_conversion(self):
+	def _update_output(self):
 		assert self._categoryName
 		assert self._inputUnitName
 		assert self._outputUnitName
@@ -868,6 +890,25 @@ class QuickConvert(object):
 		newValue = func.from_base(base, arg)
 		self._outputUnitValue.setText(str(newValue))
 
+	def _update_input(self):
+		assert self._categoryName
+		assert self._inputUnitName
+		assert self._outputUnitName
+
+		userOutput = str(self._outputUnitValue.text())
+		value = self._sanitize_value(userOutput)
+
+		unitData = unit_data.UNIT_DESCRIPTIONS[self._categoryName]
+		inputConversion, _, _ = unitData[self._inputUnitName]
+		outputConversion, _, _ = unitData[self._outputUnitName]
+
+		func, arg = outputConversion
+		base = func.to_base(value, arg)
+
+		func, arg = inputConversion
+		newValue = func.from_base(base, arg)
+		self._inputUnitValue.setText(str(newValue))
+
 	def _update_favorites(self):
 		if self._app.showFavoritesAction.isChecked():
 			assert self._categoryView.topLevelItemCount() == len(unit_data.UNIT_CATEGORIES)
@@ -947,7 +988,11 @@ class QuickConvert(object):
 
 	@misc_utils.log_exception(_moduleLogger)
 	def _on_value_edited(self, *args):
-		self._update_conversion()
+		self._update_output()
+
+	@misc_utils.log_exception(_moduleLogger)
+	def _on_output_value_edited(self, *args):
+		self._update_input()
 
 	@misc_utils.log_exception(_moduleLogger)
 	def _on_category_selection_changed(self, selected, deselected):
@@ -956,7 +1001,7 @@ class QuickConvert(object):
 			for item in self._categoryView.selectedItems()
 		]
 		assert len(selectedNames) == 1
-		self.select_category(selectedNames[0])
+		self._select_category(selectedNames[0])
 
 	@misc_utils.log_exception(_moduleLogger)
 	def _on_input_selection_changed(self, selected, deselected):
@@ -967,7 +1012,7 @@ class QuickConvert(object):
 		if selectedNames:
 			assert len(selectedNames) == 1
 			name = selectedNames[0]
-			self.select_input(name)
+			self._select_input(name)
 		else:
 			pass
 
@@ -980,7 +1025,7 @@ class QuickConvert(object):
 		if selectedNames:
 			assert len(selectedNames) == 1
 			name = selectedNames[0]
-			self.select_output(name)
+			self._select_output(name)
 		else:
 			pass
 
@@ -995,19 +1040,19 @@ class FavoritesWindow(object):
 		self._categories = QtGui.QTreeWidget()
 		self._categories.setHeaderLabels(["Categories"])
 		self._categories.setHeaderHidden(True)
+		self._categories.setSelectionMode(QtGui.QAbstractItemView.NoSelection)
 		if not IS_MAEMO:
 			self._categories.setAlternatingRowColors(True)
-		self._categories.setSelectionBehavior(QtGui.QAbstractItemView.SelectRows)
-		self._categories.setSelectionMode(QtGui.QAbstractItemView.MultiSelection)
 		self._childWidgets = []
 		for catName in self._source:
 			twi = QtGui.QTreeWidgetItem(self._categories)
 			twi.setText(0, catName)
 			self._childWidgets.append(twi)
-			if catName not in self._hidden:
-				self._categories.setItemSelected(twi, True)
-		self._selection = self._categories.selectionModel()
-		self._selection.selectionChanged.connect(self._on_selection_changed)
+			if catName in self._hidden:
+				twi.setCheckState(0, QtCore.Qt.Unchecked)
+			else:
+				twi.setCheckState(0, QtCore.Qt.Checked)
+		self._categories.itemChanged.connect(self._on_item_changed)
 
 		self._allButton = QtGui.QPushButton("All")
 		self._allButton.clicked.connect(self._on_select_all)
@@ -1080,29 +1125,36 @@ class FavoritesWindow(object):
 	@misc_utils.log_exception(_moduleLogger)
 	def _on_select_all(self, checked = False):
 		for child in self._childWidgets:
-			self._categories.setItemSelected(child, True)
+			child.setCheckState(0, QtCore.Qt.Checked)
 
 	@misc_utils.log_exception(_moduleLogger)
 	def _on_invert_selection(self, checked = False):
 		for child in self._childWidgets:
-			isSelected = self._categories.isItemSelected(child)
-			self._categories.setItemSelected(child, not isSelected)
+			state = child.checkState(0)
+			if state == QtCore.Qt.Unchecked:
+				newState = QtCore.Qt.Checked
+			elif state == QtCore.Qt.Checked:
+				newState = QtCore.Qt.Unchecked
+			else:
+				raise RuntimeError("Bad check state %r" % state)
+			child.setCheckState(0, newState)
 
 	@misc_utils.log_exception(_moduleLogger)
 	def _on_select_none(self, checked = False):
 		for child in self._childWidgets:
-			self._categories.setItemSelected(child, False)
+			child.setCheckState(0, QtCore.Qt.Unchecked)
 
 	@misc_utils.log_exception(_moduleLogger)
-	def _on_selection_changed(self, selected, deselected):
-		self._hidden.clear()
-		selectedNames = set(
-			str(item.text(0))
-			for item in self._categories.selectedItems()
-		)
-		for name in self._source:
-			if name not in selectedNames:
-				self._hidden.add(name)
+	def _on_item_changed(self, item, column):
+		state = item.checkState(column)
+		if state == QtCore.Qt.Unchecked:
+			name = str(item.text(column))
+			self._hidden.add(name)
+		elif state == QtCore.Qt.Checked:
+			name = str(item.text(column))
+			self._hidden.remove(name)
+		else:
+			raise RuntimeError("Bad check state %r" % state)
 
 	@misc_utils.log_exception(_moduleLogger)
 	def _on_close_window(self, checked = True):
@@ -1216,16 +1268,13 @@ class CategoryWindow(object):
 		self._window.close()
 
 	def select_category(self, categoryName):
-		for child in self.walk_children():
-			child.window.destroyed.disconnect(self._on_child_close)
-			child.close()
-		self._unitWindow = UnitWindow(self._window, categoryName, self._app)
-		self._unitWindow.window.destroyed.connect(self._on_child_close)
+		self._select_category(categoryName)
 
 		i = unit_data.UNIT_CATEGORIES.index(categoryName)
 		rootIndex = self._categories.rootIndex()
 		currentIndex = self._categories.model().index(i, 0, rootIndex)
 		self._categories.scrollTo(currentIndex)
+		self._categories.setItemSelected(self._categories.topLevelItem(i), True)
 		return self._unitWindow
 
 	def set_fullscreen(self, isFullscreen):
@@ -1236,6 +1285,13 @@ class CategoryWindow(object):
 		for child in self.walk_children():
 			child.set_fullscreen(isFullscreen)
 
+	def _select_category(self, categoryName):
+		for child in self.walk_children():
+			child.window.destroyed.disconnect(self._on_child_close)
+			child.close()
+		self._unitWindow = UnitWindow(self._window, categoryName, self._app)
+		self._unitWindow.window.destroyed.connect(self._on_child_close)
+
 	def _update_favorites(self):
 		if self._app.showFavoritesAction.isChecked():
 			assert self._categories.topLevelItemCount() == len(unit_data.UNIT_CATEGORIES)
@@ -1299,6 +1355,11 @@ class UnitData(object):
 	VALUE_COLUMN_1 = 2
 	UNIT_COLUMN = 3
 
+	__slots__ = [
+		"_name", "_unit", "_description", "_conversion",
+		"_value", "_integerDisplay", "_fractionalDisplay",
+	]
+
 	def __init__(self, name, unit, description, conversion):
 		self._name = name
 		self._unit = unit
@@ -1341,6 +1402,10 @@ class UnitModel(QtCore.QAbstractItemModel):
 		super(UnitModel, self).__init__(parent)
 		self._categoryName = categoryName
 		self._unitData = unit_data.UNIT_DESCRIPTIONS[self._categoryName]
+		if self._categoryName == "Computer Numbers":
+			self._sanitize_value = self._sanitize_alpha_value
+		else:
+			self._sanitize_value = self._sanitize_numeric_value
 
 		self._children = []
 		for key in unit_data.get_units(self._categoryName):
@@ -1357,19 +1422,20 @@ class UnitModel(QtCore.QAbstractItemModel):
 
 	@misc_utils.log_exception(_moduleLogger)
 	def data(self, index, role):
-		if not index.isValid():
-			return None
+		#if not index.isValid():
+		#	return None
+
+		if role == QtCore.Qt.DisplayRole:
+			item = index.internalPointer()
+			if isinstance(item, UnitData):
+				return item.data(index.column())
+			elif item is UnitData.HEADERS:
+				return item[index.column()]
 		elif role == QtCore.Qt.TextAlignmentRole:
 			return UnitData.ALIGNMENT[index.column()]
-		elif role != QtCore.Qt.DisplayRole:
+		else:
 			return None
 
-		item = index.internalPointer()
-		if isinstance(item, UnitData):
-			return item.data(index.column())
-		elif item is UnitData.HEADERS:
-			return item[index.column()]
-
 	@misc_utils.log_exception(_moduleLogger)
 	def sort(self, column, order = QtCore.Qt.AscendingOrder):
 		self._sortSettings = column, order
@@ -1400,11 +1466,10 @@ class UnitModel(QtCore.QAbstractItemModel):
 
 	@misc_utils.log_exception(_moduleLogger)
 	def index(self, row, column, parent):
-		if not self.hasIndex(row, column, parent):
-			return QtCore.QModelIndex()
-
-		if parent.isValid():
-			return QtCore.QModelIndex()
+		#if not self.hasIndex(row, column, parent):
+		#	return QtCore.QModelIndex()
+		#elif parent.isValid():
+		#	return QtCore.QModelIndex()
 
 		parentItem = UnitData.HEADERS
 		childItem = self._children[row]
@@ -1454,8 +1519,6 @@ class UnitModel(QtCore.QAbstractItemModel):
 		func, arg = self._children[fromIndex].conversion
 		base = func.to_base(value, arg)
 		for i, child in enumerate(self._children):
-			if i == fromIndex:
-				continue
 			func, arg = child.conversion
 			newValue = func.from_base(base, arg)
 			child.update_value(newValue)
@@ -1466,8 +1529,10 @@ class UnitModel(QtCore.QAbstractItemModel):
 		):
 			# Sort takes care of marking everything as changed
 			self.sort(*self._sortSettings)
+			return True
 		else:
 			self._values_changed()
+			return False
 
 	def __len__(self):
 		return len(self._children)
@@ -1482,17 +1547,18 @@ class UnitModel(QtCore.QAbstractItemModel):
 		bottomRight = self.createIndex(len(self._children)-1, len(UnitData.HEADERS)-1, self._children[-1])
 		self.dataChanged.emit(topLeft, bottomRight)
 
-	def _sanitize_value(self, userEntry):
-		if self._categoryName == "Computer Numbers":
-			if userEntry == '':
-				value = '0'
-			else:
-				value = userEntry
+	def _sanitize_alpha_value(self, userEntry):
+		if userEntry:
+			value = userEntry
 		else:
-			if userEntry == '':
-				value = 0.0
-			else:
-				value = float(userEntry)
+			value = '0'
+		return value
+
+	def _sanitize_numeric_value(self, userEntry):
+		if userEntry:
+			value = float(userEntry)
+		else:
+			value = 0.0
 		return value
 
 
@@ -1518,17 +1584,15 @@ class UnitWindow(object):
 		self._unitsModel = UnitModel(self._categoryName)
 		self._unitsView = QtGui.QTreeView()
 		self._unitsView.setModel(self._unitsModel)
-		self._unitsView.clicked.connect(self._on_unit_clicked)
 		self._unitsView.setUniformRowHeights(True)
 		self._unitsView.setSortingEnabled(True)
-		self._unitsView.setTextElideMode(QtCore.Qt.ElideNone)
 		self._unitsView.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
 		self._unitsView.setSelectionBehavior(QtGui.QAbstractItemView.SelectRows)
 		self._unitsView.setSelectionMode(QtGui.QAbstractItemView.SingleSelection)
+		self._unitsView.setHeaderHidden(True)
+		self._unitsView.clicked.connect(self._on_unit_clicked)
 		if not IS_MAEMO:
 			self._unitsView.setAlternatingRowColors(True)
-		if True:
-			self._unitsView.setHeaderHidden(True)
 
 		viewHeader = self._unitsView.header()
 		viewHeader.setSortIndicatorShown(True)
@@ -1541,10 +1605,10 @@ class UnitWindow(object):
 		viewHeader.setStretchLastSection(False)
 
 		# Trying to make things faster by locking in the initial size of the immutable columns
-		nameSize = min(viewHeader.sectionSize(UnitData.NAME_COLUMN), 125)
+		nameSize = min(viewHeader.sectionSize(UnitData.NAME_COLUMN), 150)
 		viewHeader.setResizeMode(UnitData.NAME_COLUMN, QtGui.QHeaderView.Fixed)
 		viewHeader.resizeSection(UnitData.NAME_COLUMN, nameSize)
-		unitSize = min(viewHeader.sectionSize(UnitData.UNIT_COLUMN), 125)
+		unitSize = min(viewHeader.sectionSize(UnitData.UNIT_COLUMN), 150)
 		viewHeader.setResizeMode(UnitData.UNIT_COLUMN, QtGui.QHeaderView.Fixed)
 		viewHeader.resizeSection(UnitData.UNIT_COLUMN, unitSize)
 
@@ -1683,6 +1747,9 @@ class UnitWindow(object):
 		index = self._unitsModel.index_unit(unitName)
 		self._select_unit(index)
 
+		qindex = self._unitsModel.createIndex(index, 0, self._unitsModel.get_unit(index))
+		self._unitsView.scrollTo(qindex)
+
 	def walk_children(self):
 		if self._favoritesWindow is not None:
 			yield self._favoritesWindow
@@ -1758,9 +1825,10 @@ class UnitWindow(object):
 
 	@misc_utils.log_exception(_moduleLogger)
 	def _on_value_edited(self, *args):
-		userInput = self._selectedUnitValue.text()
-		self._unitsModel.update_values(self._selectedIndex, str(userInput))
-		self._update_favorites()
+		userInput = str(self._selectedUnitValue.text())
+		orderChanged = self._unitsModel.update_values(self._selectedIndex, userInput)
+		if orderChanged:
+			self._update_favorites()
 
 	def _select_unit(self, index):
 		unit = self._unitsModel.get_unit(index)
@@ -1769,15 +1837,16 @@ class UnitWindow(object):
 		self._selectedUnitSymbol.setText(unit.unit)
 
 		self._selectedIndex = index
-		qindex = self._unitsModel.createIndex(index, 0, self._unitsModel.get_unit(index))
-		self._unitsView.scrollTo(qindex)
 		self._app.add_recent(self._categoryName, self._unitsModel.get_unit(index).name)
 
 
 def run_gonvert():
 	app = QtGui.QApplication([])
 	handle = Gonvert(app)
-	return app.exec_()
+	if constants.PROFILE_STARTUP:
+		return 0
+	else:
+		return app.exec_()
 
 
 if __name__ == "__main__":
